from configparser import SafeConfigParser, NoOptionError

from jolt import filesystem as fs
from jolt import utils


location = fs.path.join(fs.path.expanduser("~"), ".config", "jolt", "config")

_file = SafeConfigParser()
_file.read(location)
if not _file.has_section("jolt"):
    _file.add_section("jolt")

def get(section, key, default=None, expand=True):
    try:
        value = _file.get(section, key)
        return utils.expand(value) if expand else value
    except NoOptionError:
        return default

def getint(section, key, default=None):
    try:
        return _file.getint(section, key)
    except NoOptionError:
        return default

def getsize(section, key, default=None):
    units = {"B": 1, "K": 1024, "M": 1024**2, "G": 1024**3, "T": 1024**4}
    try:
        value = _file.get(section, key)
        value = value.strip()
        value = value.split()
        if len(value) == 1 and value[0][-1] in units:
            size, unit = value[0][:-1], value[0][-1]
        else:
            assert len(value) == 2, "invalid size format for {0}.{1}, "\
                "expected '<size> <unit>'"\
                .format(section, key)
            size, unit = value[0], value[1]
        assert unit in units, "invalid unit requested for {0}.{1}"\
            .format(section, key)
        return int(size)*units[unit]
    except NoOptionError:
        return default

def getfloat(section, key, default=None):
    try:
        return _file.getfloat(section, key)
    except NoOptionError:
        return default

def getboolean(section, key, default=None):
    try:
        return _file.getboolean(section, key)
    except NoOptionError:
        return default

def set(section, key, value):
    if not _file.has_section(section):
        _file.add_section(section)
    _file.set(section, key, value)

def load(file):
    _file.read(file)

def save():
    fs.makedirs(fs.path.dirname(location))
    with open(location, 'wb') as configfile:
        _file.write(configfile)

def sections():
    return _file.sections()



import hashlib
import inspect
import copy
from contextlib import contextmanager
import unittest as ut
import functools as ft
import types

from jolt import utils
from jolt.cache import *
from jolt.tools import Tools
from jolt.influence import TaskSourceInfluence


class Parameter(object):
    """ Generic task parameter type. """

    def __init__(self, default=None, values=None, required=True, help=None):
        """
        Creates a new parameter.

        Args:
            default (str, optional): An optional default value.
            values (list, optional): A list of accepted values. An
                assertion is raised if an unlisted value is assigned to the parameter.
            required (boolean, optional): If required, the parameter must be assigned
                a value before the task can be executed. The default is ``True``.
            help (str, optional): Documentation for the parameter.
                This text is displayed when running the ``info`` command on the
                associated task.

        Raises:
            AssertionError: If the parameter is assigned an illegal value
                during task execution.
        """

        self._default = default
        self._value = default
        self._accepted_values = values
        self._required = required
        self.__doc__ = help
        if default:
            self._validate(default)

    def __str__(self):
        """ Returns the parameter value as a string """
        return str(self._value) if self._value is not None else ''

    def _validate(self, value):
        assert self._accepted_values is None or value in self._accepted_values, \
            "illegal value '{0}' assigned to parameter"\
            .format(value)

    def get_default(self):
        """ Get the default value of the parameter.

        Returns:
            The default value or None if no default was given.
        """
        return self._default

    def set_default(self, value):
        """ Set the default value of the parameter.

        Args:
            value (str): The new default value of the parameter.
        """
        self._validate(value)
        if self.is_default():
            self._value = value
        self._default = value

    def is_required(self):
        """ Check if the parameter must be set to a value.

        Returns:
            True if the parameter must be assigned a value, False otherwise.
        """
        return self._required

    def is_default(self):
        """ Check if parameter is set to its default value.

        Returns:
            True if the assigned value is the default value.
        """
        return self._default == self._value

    def is_unset(self):
        """ Check if the parameter is unset.

        Returns:
            True if the assigned value is None.
        """
        return self._value is None

    def is_set(self):
        """ Check if the parameter is set to a non-default value.

        Returns:
            True if the assigned value is not the default value.
        """
        return not self.is_unset() and not self.is_default()

    def get_value(self):
        """ Get the parameter value. """
        return self._value

    def set_value(self, value):
        """ Set the parameter value.

        Args:
            value (str): The new parameter value.

        Raises:
            AssertionError: If the parameter is assigned an illegal value.
        """
        self._validate(value)
        self._value = value


class BooleanParameter(Parameter):
    """ Boolean task parameter type. """

    def __init__(self, default=None, required=True, help=None):
        """
        Creates a new parameter.

        Args:
            default (boolean, optional): An optional default boolean value.
            required (boolean, optional): If required, the parameter must be assigned
                a value before the task can be executed. The default is ``True``.
            help (str, optional): Documentation for the parameter.
                This text is displayed when running the ``info`` command on the
                associated task.

        Raises:
            AssertionError: If the parameter is assigned an illegal value
                during task execution.
        """
        default = str(default).lower() if default is not None else None
        super(BooleanParameter, self).__init__(
            default, values=["false", "true", "0", "1"],
            required=required, help=help)

    def _validate(self, value):
        assert self._accepted_values is None or value in self._accepted_values, \
            "illegal value '{0}' assigned to boolean parameter"\
            .format(value)

    def set_value(self, value):
        """ Set the parameter value.

        Args:
            value (boolean): The new parameter value. Accepted values are:
                False, True, "false, and "true", 0 and 1.

        Raises:
            AssertionError: If the parameter is assigned an illegal value.
        """
        value = str(value).lower()
        super(BooleanParameter, self).set_value(value)

    @property
    def is_true(self):
        """ The parameter value is True. """
        return str(self.get_value()) in ["true", "1"]

    @property
    def is_false(self):
        """ The parameter value is False. """
        return not self.is_true


class TaskRegistry(object):
    _instance = None

    def __init__(self):
        self.tasks = {}
        self.tests = {}
        self.instances = {}

    @staticmethod
    def get():
        if not TaskRegistry._instance:
            TaskRegistry._instance = TaskRegistry()
        return TaskRegistry._instance

    def add_task_class(self, cls):
        self.tasks[cls.name] = cls

    def add_test_class(self, cls):
        self.tests[cls.name] = cls

    def add_task(self, task, extra_params):
        name, params = utils.parse_task_name(task.name)
        params.update(extra_params or {})
        full_name = utils.format_task_name(name, params)
        self.instances[full_name] = task

    def get_task_class(self, name):
        return self.tasks.get(name)

    def get_test_class(self, name):
        return self.tests.get(name)

    def get_task_classes(self):
        return list(self.tasks.values())

    def get_test_classes(self):
        return list(self.tests.values())

    def get_task(self, name, extra_params=None):
        name, params = utils.parse_task_name(name)
        params.update(extra_params or {})
        full_name = utils.format_task_name(name, params)

        task = self.instances.get(full_name)
        if task:
            return task

        cls = self.tasks.get(name)
        if cls:
            task = cls(parameters=params)
            self.instances[full_name] = task
            return task

        cls = self.tests.get(name)
        if cls:
            task = _Test(cls, parameters=params)
            self.instances[full_name] = task
            return task

        assert task, "no such task: {0}".format(full_name)

    def set_default_parameters(self, task):
        name, params = utils.parse_task_name(task)

        cls = self.tasks.get(name)
        if not cls:
            cls = self.tests.get(name)
        assert cls, "no such task: {0}".format(task)
        cls._set_default_parameters(cls, params)

    def _create_parents(self, name):
        names = name.split("/")

        if len(names) <= 1:
            return

        prev = None
        for i in reversed(range(0, len(names))):
            name = "/".join(names[:i + 1])
            task = self.instances.get(name) or Task(name)
            if prev:
                task.requires += [prev.name]
            prev = task



class TaskBase(object):
    cacheable = True
    """ Whether the task produces an artifact or not. """

    def __init__(self, *args, **kwargs):
        super(TaskBase, self).__init__(*args, **kwargs)
        self.cacheable = self.__class__.cacheable
        self._identity = None

    @property
    def identity(self):
        return self._identity

    @identity.setter
    def identity(self, identity):
        self._identity = identity

    def _create_parameters(self):
        for key, param in self.__class__.__dict__.items():
            if isinstance(param, Parameter):
                param = copy.copy(param)
                setattr(self, key, param)

    def _set_parameters(self, params):
        params = params or {}
        for key, value in params.items():
            param = self.__dict__.get(key)
            if isinstance(param, Parameter):
                param.set_value(value)
                continue
            assert False, "no such parameter for task {0}: {1}".format(self.name, key)
        self._assert_required_parameters_assigned()

    @staticmethod
    def _set_default_parameters(cls, params):
        params = params or {}
        for key, value in params.items():
            param = cls.__dict__.get(key)
            if isinstance(param, Parameter):
                param.set_default(value)
                continue
            assert False, "no such parameter for task {0}: {1}".format(cls.name, key)

    def _assert_required_parameters_assigned(self):
        for key, param in self._get_parameter_objects().items():
            assert not param.is_required() or not param.is_unset(), \
                "required parameter '{0}' has not been set for '{1}'".format(key, self.name)

    def _get_parameter_objects(self, unset=False):
        return { key: getattr(self, key) for key in dir(self)
                 if isinstance(utils.getattr_safe(self, key), Parameter) }

    def _get_parameters(self, unset=False):
        return {key: param.get_value()
                for key, param in self._get_parameter_objects().items()
                if unset or not param.is_unset() }

    def _get_explicitly_set_parameters(self):
        return {key: param.get_value()
                for key, param in self._get_parameter_objects().items()
                if param.is_set() }

    def _get_properties(self):
        return {key: str(getattr(self, key))
                for key in dir(self)
                if utils.is_str(getattr(self, key)) }


class Task(TaskBase):
    #: Path to the .jolt file where the task was defined.
    joltdir = "."
    """ Path to the .jolt file where the task was defined. """

    name = None
    """ Name of the task. Derived from class name if not set. """

    requires = []
    """ List of dependencies to other tasks. """

    extends = ""
    """
    Name of extended task.

    A task with this attribute set is called an extension. An extension
    is executed in the context of the extended task, immediately after
    the extended task has executed.

    A common use-case for extensions is to produce additional artifacts
    from the output of another task. Also, for long-running tasks, it is
    sometimes beneficial to utilize the intermediate output from an extended
    task. The extension artifact can then be acquired more cheaply than if the
    extension had performed all of the work from scratch.

    An extension can only extend one other task.
    """

    fast = False
    """
    Indication of task speed.

    The information is used by the distributed execution strategy to
    optimize how tasks are scheduled. Scheduling tasks remotely is always
    associated with some overhead and sometimes it's beneficial to instead
    schedule fast tasks locally if possible.

    An extended task is only considered fast if all extensions are fast.
    """

    influence = []

    def __init__(self, parameters=None):
        super(Task, self).__init__()
        self.name = self.__class__.name
        self.tools = Tools(self, self.joltdir)
        self._create_parameters()
        self._set_parameters(parameters)
        self.influence = utils.as_list(self.__class__.influence)
        self.requires = utils.as_list(utils.call_or_return(self, self.__class__.requires))
        self.extends = utils.as_list(utils.call_or_return(self, self.__class__.extends))
        assert len(self.extends) == 1, "{0} extends multiple tasks, only one allowed".format(self.name)
        self.extends = self.extends[0]
        self.influence.append(TaskSourceInfluence("publish"))
        self.influence.append(TaskSourceInfluence("run"))
        self.influence.append(TaskSourceInfluence("unpack"))

    def _get_source(self, func):
        source, lines = inspect.getsourcelines(func)
        return "\n".join(source)

    def _get_source_functions(self):
        return [self.run, self.publish, self.unpack]

    def _get_source_hash(self):
        sha = hashlib.sha1()
        for func in self._get_source_functions():
            sha.update(self._get_source(func).encode())
        return sha.hexdigest()

    def _get_requires(self):
        try:
            return [self._get_expansion(req) for req in self.requires]
        except KeyError as e:
            assert False, "invalid macro expansion used in task {0}: {1} - "\
                "forgot to set the parameter?".format(self.name, e)

    def _get_extends(self):
        try:
            return self._get_expansion(self.extends)
        except KeyError as e:
            assert False, "invalid macro expansion used in task {0}: {1} - "\
                "forgot to set the parameter?".format(self.name, e)

    def _get_expansion(self, string, *args, **kwargs):
        try:
            kwargs.update(**self._get_parameters())
            kwargs.update(**self._get_properties())
            return utils.expand(string, *args, **kwargs)
        except KeyError as e:
            assert False, "invalid macro expansion used in task {0}: {1} - "\
                "forgot to set the parameter?".format(self.name, e)
    @property
    def canonical_name(self):
        return self.name.replace("/", "_")

    def is_cacheable(self):
        return self.cacheable

    def is_runnable(self):
        return True

    def info(self, fmt, *args, **kwargs):
        """
        Log information about the task.
        """
        fmt = self.tools.expand(fmt, *args, **kwargs)
        log.info(fmt, *args, **kwargs)

    def warn(self, fmt, *args, **kwargs):
        """ Log a warning concerning the task """
        fmt = self.tools.expand(fmt, *args, **kwargs)
        log.warn(fmt, *args, **kwargs)

    def error(self, fmt, *args, **kwargs):
        """ Log an error concerning the task """
        fmt = self.tools.expand(fmt, *args, **kwargs)
        log.error(fmt, *args, **kwargs)

    def run(self, deps, tools):
        """
        Performs the work of the task.

        Dependencies specified with "requires" are passed as the
        deps dictionary. The tools argument provides a set of low
        level tool functions that may be useful.

        .. code-block:: python

          with tools.cwd("path/to/subdir"):
              tools.run("make {target}")

        When using methods from the toolbox, task parameters, such
        as ``target`` above,  are automatically expanded to their values.
        """
        pass

    def publish(self, artifact, tools):
        """
        Publishes files produced by :func:`~run`.

        Files can be collected in to the artifact by calling
        artifact.collect().

        Additional metadata can be provided, such as environment
        variables that should be set whenever the task artifact is
        consumed. Example:

        .. code-block:: python

          # Append <artifact-path>/bin to the PATH
          artifact.environ.PATH.append("bin")

          # Pass an arbitrary string to a consumer
          artifact.strings.foo = "bar"

        """
        pass

    def unpack(self, artifact, tools):
        """
        Unpacks files published by publish() .

        The intention of this hook is to make necessary adjustments
        to artifact files and directories once they have been downloaded
        into the local cache on a different machine. For example,
        paths may have to be adjusted or an installer executed.

        This hook is executed in the context of a consuming task.
        """
        raise NotImplementedError()


class Resource(Task):
    """
    A resource task.

    Resources are special tasks executed in the :class:`~jolt.Context` of other tasks.
    They are invoked to acquire and release a resource, such as hardware equipment,
    before and after the execution of a task. No artifact is produced by a resource.

    Implementors should override :func:`~acquire` and :func:`~release`.

    """

    cacheable = False

    def __init__(self, *args, **kwargs):
        super(Resource, self).__init__(*args, **kwargs)
        self.influence.append(TaskSourceInfluence("acquire"))
        self.influence.append(TaskSourceInfluence("release"))

    def is_runnable(self):
        return False

    def info(self, fmt, *args, **kwargs):
        pass

    def acquire(self, artifact, deps, tools):
        """ Called to acquire the resource.

        An implementor overrides this method in a subclass. The acquired
        resource must be released manually if an exception occurs before the
        method has returned.

        Args:
            artifact (:class:`~jolt.Artifact`): The artifact associated with the resource.
                It is not possible to publish files from a resource, but the implementor
                can still use the resource to pass information to consuming tasks.
            deps (:class:`~jolt.Context`): Task execution context used to access the
                artifacts of dependencies.
            tools (:class:`~jolt.Tools`): A task specific toolbox.

        """
        pass

    def release(self, artifact, deps, tools):
        """ Called to release the resource.

        An implementor overrides this method in a subclass.

        Args:
            artifact (:class:`~jolt.Artifact`): The artifact associated with the resource.
                It is not possible to publish files from a resource, but the implementor
                can still use the resource to pass information to consuming tasks.
            deps (:class:`~jolt.Context`): Task execution context used to access the
                artifacts of dependencies.
            tools (:class:`~jolt.Tools`): A task specific toolbox.
        """
        pass

    def run(self, env, tools):
        self._run_env = env


class TaskException(Exception):
    def __init__(self, *args, **kwargs):
        super(TaskException, self).__init__(*args, **kwargs)


class _Test(Task):
    def __init__(self, test_cls, *args, **kwargs):
        self.test_cls = test_cls
        self.__class__.name = test_cls.name
        self.__class__.joltdir = test_cls.joltdir
        self.__class__.requires = test_cls.requires
        self.__class__.influence = test_cls.influence
        super(_Test, self).__init__(*args, **kwargs)
        self.influence.append(TaskSourceInfluence("setup", self.test_cls))
        self.influence.append(TaskSourceInfluence("cleanup", self.test_cls))
        for name in self._get_test_names():
            self.influence.append(TaskSourceInfluence(name, self.test_cls))

    @property
    def identity(self):
        return self.test_cls.identity

    @identity.setter
    def identity(self, identity):
        self.test_cls.identity = identity

    def _create_parameters(self):
        for key, param in self.test_cls.__dict__.items():
            if isinstance(param, Parameter):
                param = copy.copy(param)
                setattr(self, key, param)

    def _get_test_names(self):
        return [attrib for attrib in dir(self.test_cls)
                if attrib.startswith("test_")]

    def run(self, deps, tools):
        testsuite = ut.TestSuite()
        for test in self._get_test_names():
            testsuite.addTest(self.test_cls(
                test, parameters=self._get_parameters(), deps=deps, tools=tools))
        self.testresult = ut.TextTestRunner(verbosity=2).run(testsuite)
        assert self.testresult.wasSuccessful(), "tests failed"


class Test(ut.TestCase, TaskBase):
    joltdir = "."
    name = None
    requires = []
    extends = ""
    influence = []

    def __init__(self, method="runTest", parameters=None, deps=None, tools=None, *args, **kwargs):
        ut.TestCase.__init__(self, method)
        TaskBase.__init__(self)
        self.deps = deps
        self.tools = tools
        self.influence = utils.as_list(self.__class__.influence)
        self.requires = utils.as_list(utils.call_or_return(self, self.__class__.requires))
        self.extends = utils.as_list(utils.call_or_return(self, self.__class__.extends))
        assert len(self.extends) == 1, "{0} extends multiple tasks, only one allowed".format(self.name)
        self.extends = self.extends[0]
        self.name = self.__class__.name
        self._create_parameters()
        self._set_parameters(parameters)

    def setUp(self):
        self.setup(self.deps, self.tools)

    def tearDown(self):
        self.cleanup()

    def setup(self, deps, tools):
        """ Implement this method to make preparations before a test """

    def cleanup(self):
        """ Implement this method to clean up after a test """


@ArtifactAttributeSetProvider.Register
class ResourceAttributeSetProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        pass

    def parse(self, artifact, content):
        pass

    def format(self, artifact, content):
        pass

    def apply(self, task, artifact):
        task = artifact.get_task()
        if isinstance(task, Resource):
            deps = task._run_env
            deps.__enter__()
            task.acquire(artifact, deps, artifact.tools)

    def unapply(self, task, artifact):
        task = artifact.get_task()
        if isinstance(task, Resource):
            env = task._run_env
            task.release(artifact, env, artifact.tools)
            env.__exit__(None, None, None)



import shutil
import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError, ConnectTimeout, ReadTimeout
import keyring
import getpass
from ftplib import FTP, FTP_TLS


from jolt import utils
from jolt import cache
from jolt import log
from jolt import config
from jolt import filesystem as fs
from jolt.tools import Tools

NAME = "ftp"
CONNECT_TIMEOUT = 3.5


def catch(func, *args, **kwargs):
    try:
        val = func(*args, **kwargs)
        return val if val is not None else True
    except:
        return False


class FtpStorage(cache.StorageProvider):
    def __init__(self, cache):
        super(FtpStorage, self).__init__()
        self._cache = cache
        self._uri = config.get(NAME, "host")
        assert self._uri, "ftp URI not configured"
        self._path = config.get(NAME, "path", "")
        self._upload = config.getboolean(NAME, "upload", True)
        self._download = config.getboolean(NAME, "download", True)
        self._tls = config.getboolean(NAME, "tls", False)
        self._disabled = False

    def _get_auth(self):
        service = config.get(NAME, "keyring.service")
        if not service:
            return None, None

        username = config.get(NAME, "keyring.username")
        if not username:
            username = raw_input(NAME + " username: ")
            assert username, "no username configured for " + NAME
            config.set(NAME, "keyring.username", username)
            config.save()

        password = config.get(NAME, "keyring.password") or \
                   keyring.get_password(NAME, username)
        if not password:
            password = getpass.getpass(NAME + " password: ")
            assert password, "no password in keyring for " + NAME
            keyring.set_password(service, username, password)
        return username, password

    def _get_ftp(self):
        try:
            username, password = self._get_auth()
            if self._tls:
                ftp = FTP_TLS(self._uri, timeout=CONNECT_TIMEOUT)
            else:
                ftp = FTP(self._uri, timeout=CONNECT_TIMEOUT)
            ftp.login(username, password)
            if self._tls:
                ftp.prot_d()
            if not catch(ftp.cwd, self._path):
                if self._path.startswith("/"):
                    ftp.cwd("/")
                components = self._path.split("/")
                for component in components:
                    if not catch(ftp.cwd, component):
                        ftp.mkd(component)
                        ftp.cwd(component)
            return ftp
        except:
            log.exception()
            log.warn("[FTP] failed to establish server connection, disabled")
            self._disabled = True
        return None

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def download(self, node, force=False):
        if self._disabled:
            return False
        if not self._download and not force:
            return False
        with self._cache.get_artifact(node) as artifact:
            ftp = self._get_ftp()
            if ftp is None:
                return False
            pathname = artifact.get_archive_path()
            name = fs.path.basename(pathname)
            try:
                ftp.cwd(node.canonical_name)
                size = ftp.size(name)
            except:
                return False
            with log.progress("Downloading {0}".format(name), size, "B") as pbar:
                with open(pathname, 'wb') as out_file:
                    def _write(block):
                        out_file.write(block)
                        pbar.update(len(block))
                    return catch(ftp.retrbinary,
                                 "RETR {filename}".format(filename=name),
                                 callback=_write)
        return False

    def download_enabled(self):
        return not self._disabled and self._download

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def upload(self, node, force=False):
        if self._disabled:
            return True
        if not self._upload and not force:
            return True
        with self._cache.get_artifact(node) as artifact:
            ftp = self._get_ftp()
            if ftp is None:
                return False
            pathname = artifact.get_archive_path()
            name = fs.path.basename(pathname)
            size = Tools().file_size(pathname)
            with log.progress("Uploading {0}".format(name), size, "B") as pbar:
                with open(pathname, 'rb') as in_file:
                    if not catch(ftp.cwd, node.canonical_name):
                        if not catch(ftp.mkd, node.canonical_name):
                            return False
                        if not catch(ftp.cwd, node.canonical_name):
                            return False

                    catch(ftp.delete, name)

                    return catch(ftp.storbinary,
                                 "STOR {filename}".format(filename=name),
                                 in_file,
                                 callback=lambda b: pbar.update(len(b)))
        return False

    def upload_enabled(self):
        return not self._disabled and self._upload

    @utils.retried.on_exception((ConnectionError, ReadTimeout))
    def location(self, node):
        if self._disabled:
            return False
        with self._cache.get_artifact(node) as artifact:
            ftp = self._get_ftp()
            if ftp is None:
                return False
            username, _ = self._get_auth()
            username = username + "@" if username is not None else ""
            pathname = artifact.get_archive_path()
            name = fs.path.basename(pathname)
            if not catch(ftp.cwd, node.canonical_name):
                return False
            try:
                if ftp.size(name) is not None:
                    url = "ftp://{user}{uri}/{path}/{taskname}/{archive}".format(
                        user=username,
                        uri=self._uri,
                        path=self._path,
                        taskname=node.canonical_name,
                        archive=name)
                    log.hysterical("[FTP] {0}", url)
                    return url
            except:
                return False
        return False


@cache.RegisterStorage
class FtpStorageFactory(cache.StorageProviderFactory):
    @staticmethod
    def create(cache):
        log.verbose("Ftp loaded")
        return FtpStorage(cache)



from jolt.cache import *


class StringVariable(ArtifactStringAttribute):
    def __init__(self, name):
        super(StringVariable, self).__init__(name)
        self._old_value = None

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass


class StringVariableSet(ArtifactAttributeSet):
    def __init__(self):
        super(StringVariableSet, self).__init__()

    def create(self, name):
        return StringVariable(name)


@ArtifactAttributeSetProvider.Register
class StringVariableSetProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "strings", StringVariableSet())

    def parse(self, artifact, content):
        if "strings" not in content:
            return

        for key, value in content["strings"].items():
            setattr(artifact.strings, key, value)

    def format(self, artifact, content):
        if "strings" not in content:
            content["strings"] = {}

        for key, value in artifact.strings.items():
            content["strings"][key] = str(value)

    def apply(self, task, artifact):
        artifact.strings.apply(task, artifact)

    def unapply(self, task, artifact):
        artifact.strings.unapply(task, artifact)



from copy import copy

from jolt.influence import *
from jolt.tools import Tools
from jolt import utils


class DirectoryInfluenceProvider(HashInfluenceProvider):
    name = "Directory"
    path = "."
    pattern = "*"

    def __init__(self, path=None, pattern=None):
        self.path = path or self.__class__.path
        self.pattern = pattern or self.__class__.pattern

    def get_influence(self, task):
        try:
            with Tools(task, task.joltdir) as tools:
                path = task._get_expansion(self.path)
                with tools.cwd(path):
                    return tools.run("find -type f -name '{0}' | LC_ALL=C sort | xargs -n1 md5sum"
                                     .format(self.pattern),
                                     output=False, output_on_error=True)
        except KeyError as e:
            pass
        assert False, "failed to change directory to {0}".format(self.path)


def global_influence(path, pattern=None, cls=DirectoryInfluenceProvider):
    HashInfluenceRegistry.get().register(cls(path, pattern))


def influence(path, pattern=None, cls=DirectoryInfluenceProvider):
    def _decorate(taskcls):
        if "influence" not in taskcls.__dict__:
            taskcls.influence = copy(taskcls.influence)
        provider = cls(path, pattern)
        taskcls.influence.append(provider)
        return taskcls
    return _decorate



##############################################################################
#
# (C) 2016 - Robert Andersson - All rights reserved.
#
# This file and its contents are the property of Robert Andersson
# and may not be distributed, copied, or disclosed, in whole or in part,
# for any reason without written consent of the copyright holder.
#
##############################################################################


from build.tools import Tool
from build.tools.directory import PyBuildDirectoryCreator
from build.transform import msbuild 
from build.transform import pybuild
from os import path


class MSBuildCSCompiler(Tool):
    def transform(self, msproject, sources):
        ig = msproject.create_item_group()
        for source in sources:
            ig.create_compile(source.path)
        return ig


class MSBuildCXXCompiler(Tool):
    def __init__(self, cxx=True):
        self.cxx = cxx
        
    def transform(self, cxx_project, sources):
        ig = cxx_project.create_item_group()
        for source in sources:
            cl = ig.create_clcompile(source.path)
            if not self.cxx:
                cl.compileas = "CompileAsC"
                cl.compileaswinrt = "false"
        return ig


class MSBuildShaderCompiler(Tool):
    def transform(self, msproject, sources):
        ig = msproject.create_item_group()
        for source in sources:
            fx = ig.create_fxcompile(source.path)
            fx.shadertype = source.args['shadertype'].title()
            fx.shadermodel = source.args['shadermodel']
        return ig


class MSBuildImage(Tool):
    def transform(self, msproject, sources):
        ig = msproject.create_item_group()
        for source in sources:
            img = ig.create_image(source.path)
        return ig


class MSBuildMedia(Tool):
    def transform(self, msproject, sources):
        ig = msproject.create_item_group()
        for source in sources:
            img = ig.create_media(source.path)
        return ig


class MSBuildNoneTask(Tool):
    def transform(self, msproject, sources):
        ig = msproject.create_item_group()
        for source in sources:
            img = ig.create_nonetask(source.path)
        return ig


class MSBuildContent(Tool):
    def transform(self, msproject, sources):
        ig = msproject.create_item_group()
        for source in sources:
            img = ig.create_content(source.path)
        return ig


class MSBuildAppxManifest(Tool):
    def transform(self, msproject, sources):
        ig = msproject.create_item_group()
        for source in sources:
            appx = ig.create_appxmanifest(source.path)
        return ig


class PyBuildCXXCompiler(Tool):
    def __init__(self, cxx=False, env=None):
        self._executable = "cl.exe"
        self._output_ext = ".obj"
        self._cxx = cxx
        self._env = env
        
    def _directory(self, cxx_project, dirname):
        return PyBuildDirectoryCreator().transform(cxx_project, dirname)
        
    def _product(self, cxx_project, source_file):
        return path.join(cxx_project.output, '{}{}'.format(source_file, self._output_ext))
        
    def _cmdline(self, cxx_project, source_file):
        flags = cxx_project.cflags if not self._cxx else cxx_project.cxxflags

        return "{} /nologo {} /c /T{}{} /Gm /Fo{}".format(
            self._executable, 
            ' '.join(flags),
            'p' if self._cxx else 'c',
            source_file, 
            self._product(cxx_project, source_file))

    def _info(self, source_file):
        return ' [{}] {}'.format(self._executable.upper(), source_file)

    def transform(self, cxx_project, source_file):
        product = self._product(cxx_project, source_file.path)
        dir = self._directory(cxx_project, path.dirname(product))
        obj =  pybuild.Object(
            product, 
            self._cmdline(cxx_project, source_file.path), 
            self._info(source_file.path),
            self._env)
        cxx_project.add_source(source_file.path)
        cxx_project.add_job(obj)
        cxx_project.add_dependency(obj.product, source_file.path)
        cxx_project.add_dependency(obj.product, dir.product)
        return obj


class PyBuildCXXArchiver(Tool):
    def __init__(self, env=None):
        self._executable = 'lib.exe'
        self._output_pfx = ''
        self._output_ext = '.lib'
        self._env = env

    def _directory(self, cxx_project, dirname):
        return PyBuildDirectoryCreator().transform(cxx_project, dirname)

    def _product(self, cxx_project):
        return path.join(cxx_project.output, '{}{}{}'.format(self._output_pfx, cxx_project.name, self._output_ext))

    def _filelist(self, cxx_project):
        return self._product(cxx_project) + ".objects"

    def _cmdline(self, cxx_project, object_files):
        return "{} /nologo /out:{} @{}".format(
            self._executable,
            self._product(cxx_project), 
            self._filelist(cxx_project))

    def _info(self, cxx_project):
        return ' [{}] {}'.format(self._executable.upper(), cxx_project.name)

    def transform(self, project, cxx_project, object_files):
        product = self._product(cxx_project)
        dir = self._directory(cxx_project, path.dirname(product))
        library = pybuild.Object(
            product, 
            self._cmdline(cxx_project, object_files), 
            self._info(cxx_project),
            self._env)
        filelist = cxx_project.add_filelist(self._filelist(cxx_project), object_files)
        cxx_project.add_job(library)
        cxx_project.add_dependency(library.product, dir.product)
        cxx_project.add_dependency(library.product, filelist.product)
        cxx_project.add_dependency(filelist.product, dir.product)
        for obj in object_files:
            cxx_project.add_dependency(library.product, obj)
        return library


class PyBuildCXXLinker(Tool):
    def __init__(self, env=None):
        self._executable = 'link.exe'
        self._output_ext = '.exe'
        self._output_ext_dll = '.dll'
        self._env = env

    def _directory(self, cxx_project, dirname):
        return PyBuildDirectoryCreator().transform(cxx_project, dirname)

    def _product(self, project, cxx_project):
        return path.join(cxx_project.output, '{}{}'.format(
            cxx_project.name, self._output_ext_dll if hasattr(project, "shared") and project.shared else self._output_ext))

    def _filelist(self, project, cxx_project):
        return self._product(project, cxx_project) + ".objects"

    def _cmdline(self, project, cxx_project, object_files):
        libraries = ['{lib}.lib'.format(output=cxx_project.toolchain.attributes.output, lib=lib) for lib in cxx_project.libraries]
        flags = cxx_project.linkflags

        return "{} /nologo {} {} {} @{} /out:{}".format(
            self._executable, 
            '/dll' if hasattr(project, "shared") and project.shared else '',
            ' '.join(libraries),
            ' '.join(flags),
            self._filelist(project, cxx_project),
            self._product(project, cxx_project))

    def _info(self, cxx_project):
        return ' [{}] {}'.format(self._executable.upper(), cxx_project.name)

    def transform(self, project, cxx_project, object_files):
        product = self._product(project, cxx_project)
        dir = self._directory(cxx_project, path.dirname(product))
        executable = pybuild.Object(
            product, 
            self._cmdline(project, cxx_project, object_files), 
            self._info(cxx_project),
            self._env)
        filelist = cxx_project.add_filelist(self._filelist(project, cxx_project), object_files)
        cxx_project.add_job(executable)                            
        cxx_project.add_dependency(executable.product, dir.product)
        cxx_project.add_dependency(executable.product, filelist.product)
        cxx_project.add_dependency(filelist.product, dir.product)
        for obj in object_files:
            cxx_project.add_dependency(executable.product, obj)
        return executable




##############################################################################
#
# (C) 2016 - Robert Andersson - All rights reserved.
#
# This file and its contents are the property of Robert Andersson
# and may not be distributed, copied, or disclosed, in whole or in part,
# for any reason without written consent of the copyright holder.
#
##############################################################################


import sys
import os
import imp


class DepfileParser:
    def __init__(self, filename):
        self.data = ""
        self.product = ""
        self.dependencies = []

        if not os.path.exists(filename):
            return

        with open(filename) as f:
            self.data = f.read()

        self.data = self.data.replace("\n", "")
        self.data = self.data.replace("\r", "")
        self.data = self.data.replace("\\", "")

        index = self.data.find(":")
        if index < 0:
            return

        self.data = self.data[index+1:]
        self.product = self.data[0:index]
        self.dependencies = [dep for dep in self.data.split(" ") if dep]
        self.dependencies = [os.path.normpath(dep) for dep in self.dependencies]



import platform

from jolt.tasks import *
from jolt import influence

import build
import build.model
import toolchains.pam


def verbose():
    build.verbose = True


@influence.attribute("features")
@influence.attribute("incpaths")
@influence.attribute("macros")
@influence.attribute("sources")
class CXXProject(Task):
    incpaths = []
    features = []
    macros = []
    sources = []
    source_influence = True

    def __init__(self, *args, **kwargs):
        super(CXXProject, self).__init__(*args, **kwargs)
        self._init_sources()
        self.macros = utils.as_list(utils.call_or_return(self, self.__class__.macros))
        self.incpaths = utils.as_list(utils.call_or_return(self, self.__class__.incpaths))
        self.features = utils.as_list(utils.call_or_return(self, self.__class__.features))
        if self.source_influence:
            for source in self.sources:
                self.influence.append(influence.FileInfluence(source))

    def _init_sources(self):
        sources = utils.as_list(utils.call_or_return(self, self.__class__.sources))
        self.sources = []
        for l in map(self.tools.glob, sources):
            self.sources += l
        self.sources.sort()
        assert self.sources, "no source files found for task {0}".format(self.name)

    def run(self, deps, tools):
        if hasattr(self, "project"):
            with tools.environ() as env:
                self.toolchain = toolchains.pam.pam_gnu_toolchain("pam-gcc", env=env)
                self.cxx_project = self.toolchain.generate(self.project)
                self.cxx_project.transform()

    def publish(self, artifact, tools):
        artifact.collect("output/pam-gcc/{0}/{0}".format(self.name), flatten=True)


@influence.attribute("shared")
class CXXLibrary(CXXProject):
    shared = False

    def __init__(self, *args, **kwargs):
        super(CXXLibrary, self).__init__(*args, **kwargs)

    def run(self, deps, tools):
        incpaths = []
        macros = []

        for name, artifact in deps.items():
            incpaths += artifact.cxxinfo.incpaths.items()
            macros += artifact.cxxinfo.macros.items()

        self.project = build.model.cxx_library(
            self.name,
            sources=self.sources,
            features=self.features,
            incpaths=self.incpaths + incpaths,
            macros=self.macros,
            shared=self.shared,
        )
        super(CXXLibrary, self).run(deps, tools)

    def publish(self, artifact, tools):
        with tools.cwd(self.cxx_project.output):
            artifact.collect("*.a", "lib/")
            artifact.collect("*.so", "lib/")
            artifact.collect("*.dll", "lib/")
        artifact.cxxinfo.libpaths.append("lib")
        artifact.cxxinfo.libraries.append(self.name)


@influence.attribute("libpaths")
@influence.attribute("libraries")
class CXXExecutable(CXXProject):
    libpaths = []
    libraries = []

    def __init__(self, *args, **kwargs):
        super(CXXExecutable, self).__init__(*args, **kwargs)
        self.libpaths = utils.as_list(utils.call_or_return(self, self.__class__.libpaths))
        self.libraries = utils.as_list(utils.call_or_return(self, self.__class__.libraries))

    def run(self, deps, tools):
        incpaths = []
        libpaths = []
        libraries = []
        macros = []

        for name, artifact in deps.items():
            incpaths += [fs.path.join(artifact.path, path)
                         for path in artifact.cxxinfo.incpaths.items()]
            libpaths += [fs.path.join(artifact.path, path)
                         for path in artifact.cxxinfo.libpaths.items()]
            libraries += artifact.cxxinfo.libraries.items()
            macros += artifact.cxxinfo.macros.items()

        self.project = build.model.cxx_executable(
            self.name,
            sources=self.sources,
            features=self.features,
            incpaths=self.incpaths + incpaths,
            libpaths=self.libpaths + libpaths,
            libraries=self.libraries + libraries,
            macros=self.macros + macros,
        )
        super(CXXExecutable, self).run(deps, tools)

    def publish(self, artifact, tools):
        with tools.cwd(self.cxx_project.output):
            if platform.system() == "Windows":
                artifact.collect(self.name + '.exe', "bin/")
            else:
                artifact.collect(self.name, "bin/")
        artifact.environ.PATH.append("bin")



from jolt.cache import *


class CppInfoVariable(ArtifactStringAttribute):
    def __init__(self, name):
        super(CppInfoVariable, self).__init__(name)

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass


class CppInfoListVariable(CppInfoVariable):
    def __init__(self, name):
        super(CppInfoListVariable, self).__init__(name)

    def append(self, value):
        if self.get_value():
            self.set_value(self.get_value() + fs.pathsep + value)
        else:
            self.set_value(value)

    def items(self):
        value = self.get_value()
        return value.split(fs.pathsep) if value is not None else []


class CppInfoDictVariable(CppInfoListVariable):
    def __init__(self, name):
        super(CppInfoDictVariable, self).__init__(name)

    def __setitem__(self, key, value=None):
        item = "{0}={1}".format(key, value) if value is not None else key
        self.append(item)


class CppInfo(ArtifactAttributeSet):
    def __init__(self):
        super(CppInfo, self).__init__()

    def create(self, name):
        if name == "incpaths":
            return CppInfoListVariable("incpaths")
        if name == "libpaths":
            return CppInfoListVariable("libpaths")
        if name == "libraries":
            return CppInfoListVariable("libraries")
        if name == "macros":
            return CppInfoDictVariable("macros")
        assert False, "no such cxxinfo attribute: {0}".format(name)


@ArtifactAttributeSetProvider.Register
class CppInfoProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "cxxinfo", CppInfo())

    def parse(self, artifact, content):
        if "cxxinfo" not in content:
            return
        for key, value in content["cxxinfo"].items():
            setattr(artifact.cxxinfo, key, value)

    def format(self, artifact, content):
        if "cxxinfo" not in content:
            content["cxxinfo"] = {}
        for key, value in artifact.cxxinfo.items():
            content["cxxinfo"][key] = str(value)

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass



from build.tools import gnu, msvc
import build.features.pybuild as pybuild
from build.transform.toolchain import *
import build.transform.pybuild as pybuildtc


def pam_gnu_toolchain(
    name,
    inherits=None,
    cflags=None,
    cxxflags=None,
    linkflags=None,
    prefix=None,
    path=None,
    sysroot=None,
    env=None
):
    if inherits:
        tc = ToolchainRegistry.find(inherits)
    else:
        tc = pybuildtc.CXXToolchain(name)
        gnu.GNUToolFactory(prefix=prefix, path=path, env=env).configure(tc)
        pybuild.GNUFeatureFactory().configure(tc)

    tc = ToolchainExtender(name, tc)
    if cflags:
        tc.add_feature(pybuild.PyBuildCustomCFlag(cflags))
    if cxxflags:
        tc.add_feature(pybuild.PyBuildCustomCXXFlag(cxxflags))
    if linkflags:
        tc.add_feature(pybuild.PyBuildCustomLinkerFlag(linkflags))
    if sysroot:
        sysroot = "--sysroot={}".format(sysroot)
        tc.add_feature(pybuild.PyBuildCustomCFlag(sysroot))
        tc.add_feature(pybuild.PyBuildCustomCXXFlag(sysroot))
        tc.add_feature(pybuild.PyBuildCustomLinkerFlag(sysroot))
    return tc


def pam_msvc_toolchain(
    name,
    inherits,
    cflags=None,
    cxxflags=None,
    linkflags=None,
):
    tc = ToolchainRegistry.find(inherits)
    tc = ToolchainExtender(name, tc)
    if cflags:
        tc.add_feature(pybuild.PyBuildCustomCFlag(cflags))
    if cxxflags:
        tc.add_feature(pybuild.PyBuildCustomCXXFlag(cxxflags))
    if linkflags:
        tc.add_feature(pybuild.PyBuildCustomLinkerFlag(linkflags))
    return tc
